import csv
import io
import logging
import os
from typing import Any, List, Optional, Tuple, Union

import requests

from . import config

LOG = logging.getLogger(__name__)

Record = Union[Tuple, List[Any]]
Records = List[Record]


def to_csv(records: Records) -> str:
    output = io.StringIO()
    write_csv(output, records)
    return output.getvalue()


def write_csv(file, records: Records):
    # TODO: do proper type conversion here to optimize for CSV input
    #  see: https://guides.tinybird.co/guide/fine-tuning-csvs-for-fast-ingestion

    writer = csv.writer(file, quoting=csv.QUOTE_MINIMAL, lineterminator="\n")

    for record in records:
        writer.writerow(record)


class Datasource:
    """
    Abstract tinybird datasource.
    """

    endpoint: str = "/v0/datasources"

    name: str
    version: Optional[int]

    def __init__(self, name, token, version: int = None, api=None) -> None:
        self.name = name
        self.token = token
        self.version = version
        self.api = (api or config.API_URL).rstrip("/") + self.endpoint

    @property
    def canonical_name(self):
        if self.version is not None:
            return f"{self.name}__v{self.version}"
        else:
            return self.name

    def append(self, records: List[Record]) -> requests.Response:
        # TODO: replicate tinybird API concepts instead of returning Response
        query = {"name": self.canonical_name, "mode": "append"}

        headers = {"Content-Type": "text/html; charset=utf-8"}
        if self.token:
            headers["Authorization"] = f"Bearer {self.token}"

        data = self.to_csv(records)

        LOG.debug(
            "appending %d records to %s via %s(%s)",
            len(records),
            self,
            self.api,
            query,
        )
        # TODO: use multipart
        return requests.post(url=self.api, params=query, headers=headers, data=data)

    @staticmethod
    def to_csv(records: List[List[Any]]):
        return to_csv(records)

    def __str__(self):
        return f"Datasource({self.canonical_name})"

    def __repr__(self):
        return self.__str__()


class FileDatasource(Datasource):
    # for debugging/development purposes

    def __init__(self, path: str):
        name = os.path.basename(path)
        super().__init__(name, None)
        self.path = path

    def append(self, records: Records) -> requests.Response:
        if records:
            with open(self.path, "a") as fd:
                write_csv(fd, records)

        response = requests.Response()
        response.status_code = 200
        return response

    def readlines(self) -> List[str]:
        with open(self.path, "r") as fd:
            return fd.readlines()



from .client import Client
from .datasource import Datasource, Record
from .pipe import Pipe, PipeError, PipeJsonData, PipeJsonResponse, PipeMetadata, PipePageIterator
from .query import OutputFormat, QueryError, SqlQuery

__all__ = [
    "Client",
    "Datasource",
    "Record",
    "Pipe",
    "PipeError",
    "PipeMetadata",
    "PipeJsonData",
    "PipeJsonResponse",
    "PipePageIterator",
    "SqlQuery",
    "QueryError",
    "OutputFormat",
]



import logging
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests

from . import config

LOG = logging.getLogger(__name__)

PipeMetadata = List[Tuple[str, str]]
PipeJsonData = List[Dict[str, Any]]


class PipeError(Exception):
    response: requests.Response

    def __init__(self, response) -> None:
        self.response = response
        self.json: Dict = response.json()
        super().__init__(self.description)

    @property
    def description(self):
        return self.json.get("error")


class PipeJsonResponse:
    response: requests.Response
    result: Dict

    def __init__(self, response):
        self.response = response
        self.result = response.json()

    @property
    def empty(self):
        return not self.result.get("data")

    @property
    def meta(self) -> PipeMetadata:
        return [(t["name"], t["type"]) for t in self.result.get("meta", [])]

    @property
    def data(self) -> PipeJsonData:
        return self.result.get("data")


PipePageIterator = Iterator[PipeJsonResponse]


class PagedPipeQuery(PipePageIterator):
    # TODO: allow passing of custom parameters

    pipe: "Pipe"

    def __init__(self, pipe: "Pipe", page_size: int = 50, start_at: int = 0):
        self.pipe = pipe
        self.limit = page_size
        self.offset = start_at

    def __iter__(self):
        return self

    def __next__(self):
        sql = f"SELECT * FROM _ LIMIT {self.limit} OFFSET {self.offset}"
        response = self.pipe.sql(sql)
        if response.empty:
            raise StopIteration()
        self.offset += self.limit
        return response


class Pipe:
    """
    Model abstraction of a tinybird Pipe.

    TODO: implement csv mode
    """

    endpoint: str = "/v0/pipes"

    name: str
    version: Optional[int]
    resource: str

    def __init__(self, name, token, version: int = None, api=None) -> None:
        super().__init__()
        self.name = name
        self.token = token
        self.version = version
        self.resource = (api or config.API_URL).rstrip("/") + self.endpoint

    @property
    def canonical_name(self):
        if self.version is not None:
            return f"{self.name}__v{self.version}"
        else:
            return self.name

    @property
    def pipe_url(self):
        return self.resource + "/" + self.canonical_name + ".json"

    def query(self, params=None) -> PipeJsonResponse:
        params = params or dict()
        if "token" not in params and self.token:
            params["token"] = self.token

        response = requests.get(self.pipe_url, params=params)

        if response.ok:
            return PipeJsonResponse(response)
        else:
            raise PipeError(response)

    def pages(self, page_size: int = 50, start_at: int = 0) -> PipePageIterator:
        return PagedPipeQuery(pipe=self, page_size=page_size, start_at=start_at)

    def sql(self, query: str) -> PipeJsonResponse:
        """
        Run an SQL query against the pipe. For example:

            pipe.sql("select count() from _")

        See https://docs.tinybird.co/api-reference/query-api.html
        """
        headers = {"Authorization": f"Bearer {self.token}"} if self.token else {}
        params = {"q": query}

        response = requests.get(self.pipe_url, headers=headers, params=params)

        if response.ok:
            return PipeJsonResponse(response)
        else:
            raise PipeError(response)

    def __str__(self):
        return f"Pipe({self.canonical_name})"

    def __repr__(self):
        return self.__str__()



import logging
import multiprocessing
import time
from queue import Empty, Queue
from typing import Optional, Tuple

import requests

from .datasource import Datasource, Records

LOG = logging.getLogger(__name__)


class StopWorker(Exception):
    """
    An exception that indicates to stop the QueueingDatasourceAppender worker.
    """

    marker = "__STOP__"

    batch: Optional[Records]

    def __init__(self, batch: Records = None):
        self.batch = batch


class QueuingDatasourceAppender:
    """
    A QueuingDatasourceAppender reads batches of records from a source Queue and appends the batches to a data
    source. Once rate limited, it waits for the instructed amount of time (or if that is not specified,
    whatever default_retry_after is set to), before appending again.

    See https://docs.tinybird.co/api-reference/api-reference.html#limits-title

    TODO: synchronize multiple appenders through a RateLimiter concurrency structure: data source share rate limiting
     across account or workspace (not sure), so running multiple separate queuing datasource appenders will lead to
     excessive rate limiting.
    """

    default_retry_after: float = 12
    wait_after_rate_limit: float = 12

    source: Queue
    destination: Datasource
    min_interval: float

    def __init__(self, source: Queue, destination: Datasource, min_interval: float = 5) -> None:
        """
        :param source: a queue that buffers records to be appended to the datasource
        :param destination: the datasource to append to
        :param min_interval: the minimal time to wait between batches
        """
        super().__init__()
        self.source = source
        self.destination = destination
        self.stopped = multiprocessing.Event()
        self.min_interval = min_interval

    def close(self):
        if self.stopped.is_set():
            return
        self.stopped.set()
        self.source.put_nowait(StopWorker.marker)

    def run(self):
        try:
            while not self.stopped.is_set():
                try:
                    then = time.time()
                    batch, error = self._do_next_batch()

                    if error is not None:
                        # TODO: make sure the batch is not dropped on error. however, if the batch is
                        #  not appendable (for example because of errors in the data), then we need to
                        #  make sure the batch is either dropped, or we try to find the record that
                        #  causes the error.

                        raise error

                    duration = time.time() - then
                    LOG.debug("processing batch took %.2f", duration)
                    if self.min_interval:
                        self.stopped.wait(self.min_interval)

                except StopWorker as e:
                    LOG.info("indicated worker shutdown, trying to flush batch")
                    if e.batch:
                        self._retry_batch(e.batch, max_retries=2)
                    return

                except Exception:
                    LOG.exception("exception while processing batch, events will be dropped")
        finally:
            LOG.info(
                "shutting down DatasourceQueueWorker, %d elements remaining",
                self.source.qsize(),
            )

    def _get_batch(self, n=None) -> Records:
        """
        Reads the next batch from the queue.

        :param n: the maximum number of items to batch (default is entire queue)
        :return: the items from the queue as Batch
        :raises StopWorker if the StopWorker.marker was retrieved from the queue
        """
        q = self.source
        item = q.get()

        if item == StopWorker.marker:
            raise StopWorker()

        result = [item]  # block until we have at least one item

        if not n:
            n = q.qsize()

        try:
            while len(result) <= n:
                item = q.get(block=False)

                if item == StopWorker.marker:
                    raise StopWorker(result)

                result.append(item)
        except Empty:
            pass

        return result

    def _parse_retry_seconds(self, response: requests.Response) -> float:
        retry = response.headers.get("Retry-After")
        if retry:
            try:
                return float(retry) + 0.5
            except ValueError as e:
                LOG.error("error while parsing Retry-After value '%s': %s", retry, e)

        return self.default_retry_after

    def _retry_batch(self, batch, max_retries=10) -> Tuple[requests.Response, bool]:
        """
        Tries to append the given batch to the datasource for max_retries amount of times. It
        only retries if the request was rate limited, and waits for a certain amount of time
        afterwards.

        :param batch: a list of records to append to the datasource
        :param max_retries: max number of retries (defaults to 10)
        :return: a tuple with the last response and a boolean flag indicating whether the request
                 was rate-limited
        """
        limited = False
        response = None

        for _ in range(max_retries):
            response = self.destination.append(batch)

            if response.ok:
                return response, limited

            if response.status_code == 429:
                wait = self._parse_retry_seconds(response)
                limited = True
                LOG.debug(
                    "rate limited by API, keeping %d records safe for %d seconds: %s",
                    len(batch),
                    wait,
                    response.text,
                )
                time.sleep(wait)
                continue

            LOG.warning(
                "unhandled error %d: %s while appending to datasource, dropping batch",
                response.status_code,
                response.text,
            )
            return response, limited

        return response, limited

    def _do_next_batch(self) -> Tuple[Records, Optional[Exception]]:
        batch = self._get_batch()

        try:
            LOG.debug(
                "processing batch of size %d into datasource %s",
                len(batch),
                self.destination.name,
            )

            response, limited = self._retry_batch(batch)

            if limited:
                # if the request was rate-limited, we'll try again after X-Ratelimit-Reset, or the
                # wait_after_rate_limit value if it is set
                try:
                    if self.wait_after_rate_limit:
                        wait = self.wait_after_rate_limit
                    else:
                        wait = float(response.headers.get("X-Ratelimit-Reset", 0))

                    LOG.info(
                        "waiting %d second until rate-limit window resets before batching again",
                        wait,
                    )
                    time.sleep(wait)
                except ValueError:
                    LOG.exception("error while parsing X-Ratelimit-Reset value")

        except Exception as e:
            return batch, e

        return batch, None



from typing import Optional

from . import config
from .datasource import Datasource
from .pipe import Pipe
from .query import OutputFormat, SqlQuery


class Client:
    """
    Tinybird HTTP client that holds the access token and provides factory methods for resources.
    """

    def __init__(self, token: str, api: str = None):
        self.api = (api or config.API_URL).lstrip("/")
        self.token = token

    def pipe(self, name: str, version: int = None) -> Pipe:
        """
        Create an object representing a pipe with the given name, e.g.,
        "localstack_dashboard_events.json"
        """
        return Pipe(name, token=self.token, version=version, api=self.api)

    def datasource(self, name: str, version: int = None) -> Datasource:
        """
        Create an object representing a datasource with a given name.
        """
        return Datasource(name, token=self.token, version=version, api=self.api)

    def sql(self, sql: str, format: Optional[OutputFormat] = None) -> SqlQuery:
        return SqlQuery(sql, format=format, token=self.token, api=self.api)



import enum
import logging
from typing import Any, Dict, List, Optional, TypedDict

import requests

from . import config

LOG = logging.getLogger(__name__)


class OutputFormat(enum.Enum):
    # https://docs.tinybird.co/api-reference/query-api.html#id6
    CSV = "CSV"
    CSVWithNames = "CSVWithNames"
    JSON = "JSON"
    TSV = "TSV"
    TSVWithNames = "TSVWithNames"
    PrettyCompact = "PrettyCompact"
    JSONEachRow = "JSONEachRow"


class QueryMetadata(TypedDict):
    name: str
    type: str


class Statistics(TypedDict):
    elapsed: float
    rows_read: int
    bytes_read: int


JsonData = Dict[str, Any]


class JsonResult(TypedDict):
    meta: List[QueryMetadata]
    data: List[JsonData]
    rows: int
    statistics: Statistics


class QueryJsonResult:
    response: requests.Response
    result: JsonResult

    def __init__(self, response: requests.Response):
        self.response = response
        self.result = response.json()

    @property
    def empty(self):
        return not self.result.get("data")

    @property
    def meta(self) -> List[QueryMetadata]:
        return self.result.get("meta")

    @property
    def data(self) -> List[JsonData]:
        return self.result.get("data")


class QueryError(Exception):
    def __init__(self, response: requests.Response) -> None:
        self.response = response
        msg = response.text
        try:
            doc = response.json()
            if doc["error"]:
                msg = doc["error"]
        except Exception:
            pass
        super().__init__(f"{response.status_code}: {msg}")


class SqlQuery:
    """
    Tinybird SQL Query. https://docs.tinybird.co/api-reference/query-api.html#get--v0-sql
    """

    endpoint: str = "/v0/sql"

    sql: str
    format: Optional[OutputFormat]

    def __init__(self, sql: str, token, format: Optional[OutputFormat] = None, api=None) -> None:
        self.sql = sql
        self.format = format or OutputFormat.JSON
        self.token = token
        self.api = (api or config.API_URL).rstrip("/") + self.endpoint

    def get(self, format: Optional[OutputFormat] = None):
        # TODO: replicate tinybird API concepts instead of returning Response
        query = {"q": self._sql_with_format(format or self.format)}

        headers = {"Content-Type": "text/html; charset=utf-8"}
        if self.token:
            headers["Authorization"] = f"Bearer {self.token}"

        LOG.debug(
            "querying %s with query: %s",
            self.api,
            query,
        )
        response = requests.get(url=self.api, params=query, headers=headers)

        if not response.ok:
            raise QueryError(response)

        return response

    def json(self) -> QueryJsonResult:
        response = self.get(OutputFormat.JSON)

        return QueryJsonResult(response)

    def _sql_with_format(self, output_format: Optional[OutputFormat] = None):
        # TODO: handle potentially already existing FORMAT string
        if not output_format:
            return self.sql
        return self.sql + f" FORMAT {output_format.value}"
